try:
  import torch
  import pytorch_lightning as pl  # noqa
  PYTORCH_LIGHTNING_IMPORT_FAILED = False
except ImportError:
  PYTORCH_LIGHTNING_IMPORT_FAILED = True


class DCLightningModule(pl.LightningModule):
  """DeepChem Lightning Module to be used with Lightning trainer.

  TODO: Add dataloader, example code and fit, once datasetmodule
  is ready
  The lightning module is a wrapper over deepchem's torch model.
  This module directly works with pytorch lightning trainer
  which runs training for multiple epochs and also is responsible
  for setting up and training models on multiple GPUs.
  https://pytorch-lightning.readthedocs.io/en/stable/api/pytorch_lightning.core.LightningModule.html?highlight=LightningModule

  Notes
  -----
  This class requires PyTorch to be installed.
  """

  def __init__(self, dc_model):
    """Create a new DCLightningModule.

    Parameters
    ----------
    dc_model: deepchem.models.torch_models.torch_model.TorchModel
      TorchModel to be wrapped inside the lightning module.
    """
    super().__init__()
    self.dc_model = dc_model

    self.pt_model = self.dc_model.model
    self.loss = self.dc_model._loss_fn

  def configure_optimizers(self):
    return self.dc_model.optimizer._create_pytorch_optimizer(
        self.pt_model.parameters(),)

  def training_step(self, batch, batch_idx):
    """Perform a training step.

    Parameters
    ----------
    batch: A tensor, tuple or list.
    batch_idx: Integer displaying index of this batch
    optimizer_idx: When using multiple optimizers, this argument will also be present.

    Returns
    -------
    loss_outputs: outputs of losses.
    """
    batch = batch.batch_list
    inputs, labels, weights = self.dc_model._prepare_batch(batch)
    if isinstance(inputs, list):
      assert len(inputs) == 1
      inputs = inputs[0]

    outputs = self.pt_model(inputs)

    if isinstance(outputs, torch.Tensor):
      outputs = [outputs]

    if self.dc_model._loss_outputs is not None:
      outputs = [outputs[i] for i in self.dc_model._loss_outputs]

    loss_outputs = self.loss(outputs, labels, weights)

    self.log(
        "train_loss",
        loss_outputs,
        on_epoch=True,
        sync_dist=True,
        reduce_fx="mean",
        prog_bar=True,
    )

    return loss_outputs



"""
Implementation of KFAC, a second order optimizer, in PyTorch.
"""
import math
from typing import Optional, Callable, Dict, Tuple, List, Union
try:
  import torch
  import torch.optim as optim
  has_pytorch = True

except ModuleNotFoundError:
  has_pytorch = False


class KFACOptimizer(optim.Optimizer):
  """"
  This class implement the second order optimizer - KFAC, which uses Kronecker factor products of inputs and the gradients to
  get the approximate inverse fisher matrix, which is used to update the model parameters. Presently this optimizer works only
  on liner and 2D convolution layers. If you want to know more details about KFAC, please check the paper [1]_ and [2]_.

  References:
  -----------
  [1] Martens, James, and Roger Grosse. Optimizing Neural Networks with Kronecker-Factored Approximate Curvature.
  arXiv:1503.05671, arXiv, 7 June 2020. arXiv.org, http://arxiv.org/abs/1503.05671.
  [2] Grosse, Roger, and James Martens. A Kronecker-Factored Approximate Fisher Matrix for Convolution Layers.
  arXiv:1602.01407, arXiv, 23 May 2016. arXiv.org, http://arxiv.org/abs/1602.01407.
  """

  def __init__(self,
               model: torch.nn.Module,
               lr: float = 0.001,
               momentum: float = 0.9,
               stat_decay: float = 0.95,
               damping: float = 0.001,
               kl_clip: float = 0.001,
               weight_decay: float = 0,
               TCov: int = 10,
               TInv: int = 100,
               batch_averaged: bool = True,
               mean: bool = False):
    """
    Parameters:
    -----------
    model: torch.nn.Module
      The model to be optimized.
    lr: float (default: 0.001)
      Learning rate for the optimizer.
    momentum: float (default: 0.9)
      Momentum for the optimizer.
    stat_decay: float (default: 0.95)
      Decay rate for the update of covariance matrix with mean.
    damping: float (default: 0.001)
      damping factor for the update of covariance matrix.
    kl_clip: float (default: 0.001)
      Clipping value for the update of covariance matrix.
    weight_decay: float (default: 0)
      weight decay for the optimizer.
    Tcov: int (default: 10)
      The number of steps to update the covariance matrix.
    Tinv: int (default: 100)
      The number of steps to calculate the inverse of covariance matrix.
    batch_averaged: bool (default: True)
      States whether to use batch averaged covariance matrix.
    mean: bool (default: False)
      States whether to use mean centered covariance matrix.
    """

    if lr < 0.0:
      raise ValueError("Invalid learning rate: {}".format(lr))
    if momentum < 0.0:
      raise ValueError("Invalid momentum value: {}".format(momentum))
    if weight_decay < 0.0:
      raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
    defaults = dict(lr=lr,
                    momentum=momentum,
                    damping=damping,
                    weight_decay=weight_decay)
    super(KFACOptimizer, self).__init__(model.parameters(), defaults)
    self.batch_averaged = batch_averaged

    self.known_modules = {'Linear', 'Conv2d'}

    self.modules: List[torch.nn.Module] = []

    self.model = model
    self._prepare_model()

    self.steps = 0

    self.m_aa: Dict[torch.nn.Module, torch.Tensor] = {}
    self.m_gg: Dict[torch.nn.Module, torch.Tensor] = {}
    self.Q_a: Dict[torch.nn.Module, torch.Tensor] = {}
    self.Q_g: Dict[torch.nn.Module, torch.Tensor] = {}
    self.d_a: Dict[torch.nn.Module, torch.Tensor] = {}
    self.d_g: Dict[torch.nn.Module, torch.Tensor] = {}
    self.stat_decay = stat_decay

    self.kl_clip = kl_clip
    self.TCov = TCov
    self.TInv = TInv

    self.mean = mean

  def try_contiguous(self, x: torch.Tensor) -> torch.Tensor:
    """
    Checks the memory layout of the input tensor and changes it to contiguous type.

    Parameters:
    -----------
    x: torch.Tensor
      The input tensor to be made contiguous in memory, if it is not so.

    Return:
    -------
    torch.Tensor
      Tensor with contiguous memory
    """
    if not x.is_contiguous():
      x = x.contiguous()

    return x

  def _extract_patches(
      self, x: torch.Tensor, kernel_size: Tuple[int, ...],
      stride: Tuple[int, ...], padding: Union[int, str,
                                              Tuple[int, ...]]) -> torch.Tensor:
    """
    Extract patches of a given size from the input tensor given. Used in calculating
    the matrices for the kronecker product in the case of 2d Convolutions.

    Parameters:
    -----------
    x: torch.Tensor
      The input feature maps. with the size of (batch_size, in_c, h, w)
    kernel_size: Tuple[int, ...]
      the kernel size of the conv filter.
    stride: Tuple[int, ...]
      the stride of conv operation.
    padding: Union[int, str, Tuple[int, ...]]
      number of paddings. be a tuple of two elements

    Return:
    -------
    torch.Tensor:
      Extracted patches with shape (batch_size, out_h, out_w, in_c*kh*kw)
    """
    if isinstance(padding, tuple):
      if padding[0] + padding[1] > 0:
        x = torch.nn.functional.pad(x, (padding[1], padding[1], padding[0],
                                        padding[0])).data  # Actually check dims
    elif isinstance(padding, int):
      if padding > 0:
        x = torch.nn.functional.pad(x,
                                    (padding, padding, padding, padding)).data
    elif isinstance(padding, str):
      if padding == 'VALID':
        pad = int((kernel_size[0] - 1) / 2)
        x = torch.nn.functional.pad(x, (pad, pad, pad, pad)).data

    x = x.unfold(2, kernel_size[0], stride[0])
    x = x.unfold(3, kernel_size[1], stride[1])
    x = x.transpose_(1, 2).transpose_(2, 3).contiguous()
    x = x.view(x.size(0), x.size(1), x.size(2),
               x.size(3) * x.size(4) * x.size(5))
    return x

  def compute_cov_a(self, a: torch.Tensor,
                    layer: torch.nn.Module) -> torch.Tensor:
    """
    Compute the covariance matrix of the A matrix (the output of each layer).

    Parameters:
    -----------
    a: torch.Tensor
      It is the output of the layer for which the covariance matrix should be calculated.
    layer: torch.nn.Module
      It specifies the type of layer from which the output of the layer is taken.

    Returns:
    --------
    torch.Tensor
      The covariance matrix of the A matrix.
    """
    if isinstance(layer, torch.nn.Linear):
      batch_size = a.size(0)
      if layer.bias is not None:
        a = torch.cat((a, a.new(a.size(0), 1).fill_(1)), 1)

    elif isinstance(layer, torch.nn.Conv2d):
      batch_size = a.size(0)
      a = self._extract_patches(a, layer.kernel_size, layer.stride,
                                layer.padding)
      spatial_size = a.size(1) * a.size(2)
      a = a.view(-1, a.size(-1))
      if layer.bias is not None:
        a = torch.cat((a, a.new(a.size(0), 1).fill_(1)), 1)
      a = a / spatial_size

    return a.t() @ (a / batch_size)

  def compute_cov_g(self, g: torch.Tensor,
                    layer: torch.nn.Module) -> torch.Tensor:
    """
    Compute the covariance matrix of the G matrix (the gradient of the layer).

    Parameters:
    -----------
    g: torch.Tensor
      It is the gradient of the layer for which the covariance matrix should be calculated.
    layer: torch.nn.Module
      It specifies the type of layer from which the output of the layer is taken.

    Returns:
    --------
    torch.Tensor
      The covariance matrix of the G matrix.
    """
    if isinstance(layer, torch.nn.Linear):
      batch_size = g.size(0)
      if self.batch_averaged:
        cov_g = g.t() @ (g * batch_size)
      else:
        cov_g = g.t() @ (g / batch_size)

    elif isinstance(layer, torch.nn.Conv2d):
      spatial_size = g.size(2) * g.size(3)
      batch_size = g.shape[0]
      g = g.transpose(1, 2).transpose(2, 3)
      g = self.try_contiguous(g)
      g = g.view(-1, g.size(-1))
      if self.batch_averaged:
        g = g * batch_size
      g = g * spatial_size
      cov_g = g.t() @ (g / g.size(0))

    return cov_g

  def _save_input(self, module: torch.nn.Module, input: torch.Tensor):
    """
    Updates the input of the layer using exponentially weighted averages of the layer input.

    Parameters:
    -----------
    module: torch.nn.Module
      specifies the layer for which the input should be taken
    input: torch.Tensor
      the input matrix which should get updated
    """
    if self.steps % self.TCov == 0:
      aa = self.compute_cov_a(input[0].data, module)
      # Initialize buffers
      if self.steps == 0:
        self.m_aa[module] = torch.diag(aa.new(aa.size(0)).fill_(1))
      self.m_aa[module] *= self.stat_decay + aa * (1 - self.stat_decay)

  def _save_grad_output(self, module: torch.nn.Module, grad_input: torch.Tensor,
                        grad_output: torch.Tensor):
    """
    Updates the backward gradient of the layer using exponentially weighted averages of the layer input.

    Parameters:
    -----------
    module: torch.nn.Module
      specifies the layer for which the gradient should be taken
    input: torch.Tensor
      the gradient matrix which should get updated
    """
    # Accumulate statistics for Fisher matrices
    if self.steps % self.TCov == 0:
      gg = self.compute_cov_g(grad_output[0].data, module)
      # Initialize buffers
      if self.steps == 0:
        self.m_gg[module] = torch.diag(gg.new(gg.size(0)).fill_(1))
      self.m_gg[module] *= self.stat_decay + gg * (1 - self.stat_decay)

  def _prepare_model(self):
    """"
    Attaches hooks(saving the ouptut and grad according to the update function) to the model for
    to calculate gradients at every step.
    """
    count = 0
    for module in self.model.modules():
      classname = module.__class__.__name__
      if classname in self.known_modules:
        self.modules.append(module)
        module.register_forward_pre_hook(self._save_input)
        module.register_backward_hook(self._save_grad_output)
        count += 1

  def _update_inv(self, m: torch.nn.Module):
    """
    Does eigen decomposition of the input(A) and gradient(G) matrix for computing inverse of the ~ fisher.

    Parameter:
    ----------
    m: torch.nn.Module
      This is the layer for which the eigen decomposition should be done on.
    """
    eps = 1e-10  # for numerical stability

    if self.mean:
      self.d_a[m], self.Q_a[m] = torch.symeig(self.m_aa[m] -
                                              torch.mean(self.m_aa[m]),
                                              eigenvectors=True)
      self.d_g[m], self.Q_g[m] = torch.symeig(self.m_gg[m] -
                                              torch.mean(self.m_gg[m]),
                                              eigenvectors=True)
    else:
      self.d_a[m], self.Q_a[m] = torch.symeig(self.m_aa[m], eigenvectors=True)
      self.d_g[m], self.Q_g[m] = torch.symeig(self.m_gg[m], eigenvectors=True)

    self.d_a[m].mul_((self.d_a[m] > eps).float())
    self.d_g[m].mul_((self.d_g[m] > eps).float())

  @staticmethod
  def _get_matrix_form_grad(m: torch.nn.Module):
    """
    Returns the gradient of the layer in a matrix form

    Parameter:
    ----------
    m: torch.nn.Module
      the layer for which the gradient must be calculated

    Return:
    -------
    torch.tensor
      a matrix form of the gradient. it should be a [output_dim, input_dim] matrix.
    """
    if isinstance(m, torch.nn.Conv2d):
      p_grad_mat = m.weight.grad.data.view(m.weight.grad.data.size(0),
                                           -1)  # n_filters * (in_c * kw * kh)
    elif isinstance(m, torch.nn.Linear):
      p_grad_mat = m.weight.grad.data
    else:
      raise NotImplementedError(
          "KFAC optimizer currently support only Linear and Conv2d layers")

    if m.bias is not None:
      if isinstance(m.bias.grad.data, torch.Tensor):
        p_grad_mat = torch.cat([p_grad_mat, m.bias.grad.data.view(-1, 1)], 1)
      else:
        raise TypeError("bias.grad.data should be a Tensor")
    return p_grad_mat

  def _get_natural_grad(self, m: torch.nn.Module, p_grad_mat: torch.Tensor,
                        damping: float) -> List[torch.Tensor]:
    """
    This function returns the product of inverse of the fisher matrix and the weights gradient.

    Parameters:
    -----------
    m: torch.nn.Module
      Specifies the layer for which the calculation must be done on.
    p_grad_mat: torch.Tensor
      the gradients in matrix form isinstance(m.weight.grad.data, torch.Tensor) and i
    damping: float
      the damping factor for the calculation

    Return:
    -------
    torch.Tensor
      the product of inverse of the fisher matrix and the weights gradient.
    """
    # p_grad_mat is of output_dim * input_dim
    # inv((ss')) p_grad_mat inv(aa') = [ Q_g (1/R_g) Q_g^T ] @ p_grad_mat @ [Q_a (1/R_a) Q_a^T]
    v1 = self.Q_g[m].t() @ p_grad_mat @ self.Q_a[m]
    v2 = v1 / (self.d_g[m].unsqueeze(1) * self.d_a[m].unsqueeze(0) + damping)
    a = self.Q_g[m] @ v2 @ self.Q_a[m].t()
    if m.bias is not None:
      # we always put gradient w.r.t weight in [0]
      # and w.r.t bias in [1]
      if isinstance(m.weight.grad.data, torch.Tensor) and isinstance(
          m.bias.grad.data, torch.Tensor):
        v = [a[:, :-1], a[:, -1:]]
        v[0] = v[0].view(m.weight.grad.data.size())
        v[1] = v[1].view(m.bias.grad.data.size())
      else:
        raise TypeError(
            "weight.grad.data and bias.grad.data should be a Tensor")
    else:
      v = [a.view(m.weight.grad.data.size())]

    return v

  def _kl_clip_and_update_grad(self, updates: Dict[torch.nn.Module,
                                                   List[torch.Tensor]],
                               lr: float):
    """
    Performs clipping on the updates matrix, if the value is large. Then final value is updated in the backwards gradient data

    Parameters:
    -----------
    updates: Dict[torch.nn.Module,List[torch.Tensor]]
      A dicitonary containing the product of gradient and fisher inverse of each layer.
    lr: float
      learning rate of the optimizer
    """
    # do kl clip
    vg_sum = 0.0
    for m in self.modules:
      v = updates[m]
      vg_sum += (v[0] * m.weight.grad.data * lr**2).sum().item()
      if m.bias is not None:
        vg_sum += (v[1] * m.bias.grad.data * lr**2).sum().item()
    nu = min(1.0, math.sqrt(self.kl_clip / vg_sum))

    for m in self.modules:
      v = updates[m]
      if isinstance(m.weight.grad.data, torch.Tensor):
        m.weight.grad.data.copy_(v[0])
        m.weight.grad.data.mul_(nu)
      else:
        raise TypeError("weight.grad.data should be a Tensor")
      if m.bias is not None:
        if isinstance(m.bias.grad.data, torch.Tensor):
          m.bias.grad.data.copy_(v[1])
          m.bias.grad.data.mul_(nu)
        else:
          raise TypeError("bias.grad.data should be a Tensor")

  def _step(self, closure: Optional[Callable] = None):
    """
    Called in every step of the optimizer, updating the model parameters from the gradient by the KFAC equation.
    Also, performs weight decay and adds momentum if any.

    Parameters:
    -----------
    closure: Callable, optional(default: None)
      an optional customizable function to be passed which can be used to clear the gradients and other compute loss for every step.
    """
    for group in self.param_groups:
      weight_decay = group['weight_decay']
      momentum = group['momentum']
      lr = group['lr']

      for p in group['params']:
        if p.grad is None:
          continue
        d_p = p.grad.data
        if weight_decay != 0 and self.steps >= 20 * self.TCov:
          d_p.add_(weight_decay, p.data)
        if momentum != 0:
          param_state = self.state[p]
          if 'momentum_buffer' not in param_state:
            buf = param_state['momentum_buffer'] = torch.zeros_like(p.data)
            buf.mul_(momentum).add_(d_p)
          else:
            buf = param_state['momentum_buffer']
            buf.mul_(momentum).add_(d_p)
          d_p = buf

        torch.add(p.data, -lr, d_p, out=p.data)

  def step(self, closure: Optional[Callable] = None):
    """
    This is the function that gets called in each step of the optimizer to update the weights and biases of the model.

    Parameters:
    -----------
    closure: Callable, optional(default: None)
      an optional customizable function to be passed which can be used to clear the gradients and other compute loss for every step.
    """
    group = self.param_groups[0]
    lr = group['lr']
    damping = group['damping']
    updates = {}
    for m in self.modules:
      if self.steps % self.TInv == 0:
        self._update_inv(m)
      p_grad_mat = self._get_matrix_form_grad(m)
      v = self._get_natural_grad(m, p_grad_mat, damping)
      updates[m] = v
    self._kl_clip_and_update_grad(updates, lr)

    self._step(closure)
    self.steps += 1



"""
Gathers all splitters in one place for convenient imports
"""
# flake8: noqa

# basic splitter
from deepchem.splits.splitters import Splitter
from deepchem.splits.splitters import RandomSplitter
from deepchem.splits.splitters import RandomStratifiedSplitter
from deepchem.splits.splitters import RandomGroupSplitter
from deepchem.splits.splitters import SingletaskStratifiedSplitter
from deepchem.splits.splitters import IndexSplitter
from deepchem.splits.splitters import SpecifiedSplitter

# molecule splitter
from deepchem.splits.splitters import ScaffoldSplitter
from deepchem.splits.splitters import MolecularWeightSplitter
from deepchem.splits.splitters import MaxMinSplitter
from deepchem.splits.splitters import FingerprintSplitter
from deepchem.splits.splitters import ButinaSplitter

# other splitter
from deepchem.splits.task_splitter import merge_fold_datasets
from deepchem.splits.task_splitter import TaskSplitter

#################################################################
# Removed API
#################################################################

import logging
logger = logging.getLogger(__name__)


class IndiceSplitter:

  def __init__(self, valid_indices=None, test_indices=None):
    raise ImportError("IndiceSplitter was renamed to SpecifiedSplitter.\n"
                      "Please use SpecifiedSplitter instead of IndiceSplitter.")



from rdkit import Chem
import numpy as np
import logging
from typing import List, Tuple, Union, Dict, Set, Sequence, Optional
from deepchem.utils.typing import RDKitAtom, RDKitMol, RDKitBond

from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.feat.graph_data import GraphData
from deepchem.feat.molecule_featurizers.circular_fingerprint import CircularFingerprint
from deepchem.feat.molecule_featurizers.rdkit_descriptors import RDKitDescriptors

from deepchem.utils.molecule_feature_utils import one_hot_encode
from deepchem.utils.molecule_feature_utils import get_atom_total_degree_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_formal_charge_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_total_num_Hs_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_hybridization_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_is_in_aromatic_one_hot

from deepchem.feat.graph_features import bond_features as b_Feats

logger = logging.getLogger(__name__)


class GraphConvConstants(object):
  """
  A class for holding featurization parameters.
  """

  MAX_ATOMIC_NUM = 100
  ATOM_FEATURES: Dict[str, List[int]] = {
      'atomic_num': list(range(MAX_ATOMIC_NUM)),
      'degree': [0, 1, 2, 3, 4, 5],
      'formal_charge': [-1, -2, 1, 2, 0],
      'chiral_tag': [0, 1, 2, 3],
      'num_Hs': [0, 1, 2, 3, 4]
  }
  ATOM_FEATURES_HYBRIDIZATION: List[str] = ["SP", "SP2", "SP3", "SP3D", "SP3D2"]
  # Dimension of atom feature vector
  ATOM_FDIM = sum(len(choices) + 1 for choices in ATOM_FEATURES.values()) + len(
      ATOM_FEATURES_HYBRIDIZATION) + 1 + 2
  # len(choices) +1 and len(ATOM_FEATURES_HYBRIDIZATION) +1 to include room for unknown set
  # + 2 at end for is_in_aromatic and mass
  BOND_FDIM = 14

  # dictionary of available feature generators
  FEATURE_GENERATORS: Dict[str, MolecularFeaturizer] = {
      "morgan":
          CircularFingerprint(radius=2, size=2048, sparse=False),
      "morgan_count":
          CircularFingerprint(radius=2,
                              size=2048,
                              sparse=False,
                              is_counts_based=True),
      "rdkit_desc":
          RDKitDescriptors(use_bcut2d=False),
      "rdkit_desc_normalized":
          RDKitDescriptors(use_bcut2d=False, is_normalized=True)
  }


def get_atomic_num_one_hot(atom: RDKitAtom,
                           allowable_set: List[int],
                           include_unknown_set: bool = True) -> List[float]:
  """Get a one-hot feature about atomic number of the given atom.

  Parameters
  ---------
  atom: RDKitAtom
    RDKit atom object
  allowable_set: List[int]
    The range of atomic numbers to consider.
  include_unknown_set: bool, default False
    If true, the index of all types not in `allowable_set` is `len(allowable_set)`.

  Returns
  -------
  List[float]
    A one-hot vector of atomic number of the given atom.
    If `include_unknown_set` is False, the length is `len(allowable_set)`.
    If `include_unknown_set` is True, the length is `len(allowable_set) + 1`.
  """
  return one_hot_encode(atom.GetAtomicNum() - 1, allowable_set,
                        include_unknown_set)


def get_atom_chiral_tag_one_hot(
    atom: RDKitAtom,
    allowable_set: List[int],
    include_unknown_set: bool = True) -> List[float]:
  """Get a one-hot feature about chirality of the given atom.

  Parameters
  ---------
  atom: RDKitAtom
    RDKit atom object
  allowable_set: List[int]
    The list of chirality tags to consider.
  include_unknown_set: bool, default False
    If true, the index of all types not in `allowable_set` is `len(allowable_set)`.

  Returns
  -------
  List[float]
    A one-hot vector of chirality of the given atom.
    If `include_unknown_set` is False, the length is `len(allowable_set)`.
    If `include_unknown_set` is True, the length is `len(allowable_set) + 1`.
  """
  return one_hot_encode(atom.GetChiralTag(), allowable_set, include_unknown_set)


def get_atom_mass(atom: RDKitAtom) -> List[float]:
  """Get vector feature containing downscaled mass of the given atom.

  Parameters
  ---------
  atom: RDKitAtom
    RDKit atom object

  Returns
  -------
  List[float]
    A vector of downscaled mass of the given atom.
  """
  return [atom.GetMass() * 0.01]


def atom_features(
    atom: RDKitAtom,
    functional_groups: List[int] = None,
    only_atom_num: bool = False) -> Sequence[Union[bool, int, float]]:
  """Helper method used to compute atom feature vector.

  Deepchem already contains an atom_features function, however we are defining a new one here due to the need to handle features specific to DMPNN.

  Parameters
  ----------
  atom: RDKitAtom
    Atom to compute features on.
  functional_groups: List[int]
    A k-hot vector indicating the functional groups the atom belongs to.
    Default value is None
  only_atom_num: bool
    Toggle to build a feature vector for an atom containing only the atom number information.

  Returns
  -------
  features: Sequence[Union[bool, int, float]]
    A list of atom features.

  Examples
  --------
  >>> import deepchem as dc
  >>> from rdkit import Chem
  >>> mol = Chem.MolFromSmiles('C')
  >>> atom = mol.GetAtoms()[0]
  >>> features = dc.feat.molecule_featurizers.dmpnn_featurizer.atom_features(atom)
  >>> type(features)
  <class 'list'>
  >>> len(features)
  133
  """

  if atom is None:
    features: Sequence[Union[bool, int,
                             float]] = [0] * GraphConvConstants.ATOM_FDIM

  elif only_atom_num:
    features = []
    features += get_atomic_num_one_hot(
        atom, GraphConvConstants.ATOM_FEATURES['atomic_num'])
    features += [0] * (
        GraphConvConstants.ATOM_FDIM - GraphConvConstants.MAX_ATOMIC_NUM - 1
    )  # set other features to zero

  else:
    features = []
    features += get_atomic_num_one_hot(
        atom, GraphConvConstants.ATOM_FEATURES['atomic_num'])
    features += get_atom_total_degree_one_hot(
        atom, GraphConvConstants.ATOM_FEATURES['degree'])
    features += get_atom_formal_charge_one_hot(
        atom, GraphConvConstants.ATOM_FEATURES['formal_charge'])
    features += get_atom_chiral_tag_one_hot(
        atom, GraphConvConstants.ATOM_FEATURES['chiral_tag'])
    features += get_atom_total_num_Hs_one_hot(
        atom, GraphConvConstants.ATOM_FEATURES['num_Hs'])
    features += get_atom_hybridization_one_hot(
        atom, GraphConvConstants.ATOM_FEATURES_HYBRIDIZATION, True)
    features += get_atom_is_in_aromatic_one_hot(atom)
    features = [int(feature) for feature in features]
    features += get_atom_mass(atom)

    if functional_groups is not None:
      features += functional_groups
  return features


def bond_features(bond: RDKitBond) -> Sequence[Union[bool, int, float]]:
  """wrapper function for bond_features() already available in deepchem, used to compute bond feature vector.

  Parameters
  ----------
  bond: RDKitBond
    Bond to compute features on.

  Returns
  -------
  features: Sequence[Union[bool, int, float]]
    A list of bond features.

  Examples
  --------
  >>> import deepchem as dc
  >>> from rdkit import Chem
  >>> mol = Chem.MolFromSmiles('CC')
  >>> bond = mol.GetBondWithIdx(0)
  >>> b_features = dc.feat.molecule_featurizers.dmpnn_featurizer.bond_features(bond)
  >>> type(b_features)
  <class 'list'>
  >>> len(b_features)
  14
  """
  if bond is None:
    b_features: Sequence[Union[
        bool, int, float]] = [1] + [0] * (GraphConvConstants.BOND_FDIM - 1)

  else:
    b_features = [0] + b_Feats(bond, use_extended_chirality=True)
  return b_features


def map_reac_to_prod(
    mol_reac: RDKitMol,
    mol_prod: RDKitMol) -> Tuple[Dict[int, int], List[int], List[int]]:
  """
  Function to build a dictionary of mapping atom indices in the reactants to the products.

  Parameters
  ----------
  mol_reac: RDKitMol
  An RDKit molecule of the reactants.

  mol_prod: RDKitMol
  An RDKit molecule of the products.

  Returns
  -------
  mappings: Tuple[Dict[int,int],List[int],List[int]]
  A tuple containing a dictionary of corresponding reactant and product atom indices,
  list of atom ids of product not part of the mapping and
  list of atom ids of reactant not part of the mapping
  """
  only_prod_ids: List[int] = []
  prod_map_to_id: Dict[int, int] = {}
  mapnos_reac: Set[int] = set(
      [atom.GetAtomMapNum() for atom in mol_reac.GetAtoms()])
  for atom in mol_prod.GetAtoms():
    mapno = atom.GetAtomMapNum()
    if (mapno > 0):
      prod_map_to_id[mapno] = atom.GetIdx()
      if (mapno not in mapnos_reac):
        only_prod_ids.append(atom.GetIdx())
    else:
      only_prod_ids.append(atom.GetIdx())
  only_reac_ids: List[int] = []
  reac_id_to_prod_id: Dict[int, int] = {}
  for atom in mol_reac.GetAtoms():
    mapno = atom.GetAtomMapNum()
    if (mapno > 0):
      try:
        reac_id_to_prod_id[atom.GetIdx()] = prod_map_to_id[mapno]
      except KeyError:
        only_reac_ids.append(atom.GetIdx())
    else:
      only_reac_ids.append(atom.GetIdx())
  mappings: Tuple[Dict[int, int], List[int],
                  List[int]] = (reac_id_to_prod_id, only_prod_ids,
                                only_reac_ids)
  return mappings


class _MapperDMPNN:
  """
  This class is a helper class for DMPNN featurizer to generate concatenated feature vector and mapping.

  `self.f_ini_atoms_bonds_zero_padded` is the concatenated feature vector which contains
  concatenation of initial atom and bond features.

  `self.mapping` is the mapping which maps bond index to 'array of indices of the bonds'
  incoming at the initial atom of the bond (excluding the reverse bonds)
  """

  def __init__(self, datapoint: RDKitMol, concat_fdim: int,
               f_atoms_zero_padded: np.ndarray):
    """
    Parameters
    ----------
    datapoint: RDKitMol
      RDKit mol object.
    concat_fdim: int
      dimension of feature vector with concatenated atom (initial) and bond features
    f_atoms_zero_padded: np.ndarray
      mapping from atom index to atom features | initial input is a zero padding
    """
    self.datapoint = datapoint
    self.concat_fdim = concat_fdim
    self.f_atoms_zero_padded = f_atoms_zero_padded

    # number of atoms
    self.num_atoms: int = len(f_atoms_zero_padded) - 1

    # number of bonds
    self.num_bonds: int = 0

    # mapping from bond index to concat(in_atom, bond) features | initial input is a zero padding
    self.f_ini_atoms_bonds_zero_padded: np.ndarray = np.asarray(
        [[0] * (self.concat_fdim)], dtype=float)

    # mapping from atom index to list of indices of incoming bonds
    self.atom_to_incoming_bonds: List[List[int]] = [
        [] for i in range(self.num_atoms + 1)
    ]

    # mapping from bond index to the index of the atom the bond is coming from
    self.bond_to_ini_atom: List[int] = [0]

    # mapping from bond index to the index of the reverse bond
    self.b2revb: List[int] = [0]

    self.mapping: np.ndarray = np.empty(0)

    self._generate_mapping()

  def _generate_mapping(self):
    """
    Generate mapping which maps bond index to 'array of indices of the bonds'
    incoming at the initial atom of the bond (reverse bonds are not considered).

    Steps:
    - Iterate such that all bonds in the mol are considered.
      For each iteration: (if bond exists)
      - Update the `self.f_ini_atoms_bonds_zero_padded` concatenated feature vector.
      - Update secondary mappings.
    - Modify `self.atom_to_incoming_bonds` based on maximum number of bonds.
    - Get mapping based on `self.atom_to_incoming_bonds` and `self.bond_to_ini_atom`.
    - Replace reverse bond values with 0
    """
    for a1 in range(1, self.num_atoms + 1):
      for a2 in range(a1 + 1, self.num_atoms + 1):
        if not self._update_concat_vector(a1, a2):
          continue
        self._update_secondary_mappings(a1, a2)
        self.num_bonds += 2
    self._modify_based_on_max_bonds()

    # get mapping which maps bond index to 'array of indices of the bonds' incoming at the initial atom of the bond
    self.mapping = np.asarray(
        self.atom_to_incoming_bonds)[self.bond_to_ini_atom]

    self._replace_rev_bonds()

  def _extend_concat_feature(self, a1: int, bond_feature: np.ndarray):
    """
    Helper method to concatenate initial atom and bond features and append them to `self.f_ini_atoms_bonds_zero_padded`.

    Parameters
    ----------
    a1: int
      index of the atom where the bond starts
    bond_feature: np.ndarray
      array of bond features
    """
    concat_input: np.ndarray = np.concatenate(
        (self.f_atoms_zero_padded[a1], bond_feature),
        axis=0).reshape([1, self.concat_fdim])
    self.f_ini_atoms_bonds_zero_padded = np.concatenate(
        (self.f_ini_atoms_bonds_zero_padded, concat_input), axis=0)

  def _update_concat_vector(self, a1: int, a2: int):
    """
    Method to update `self.f_ini_atoms_bonds_zero_padded` with features of the bond between atoms `a1` and `a2`.

    Parameters
    ----------
    a1: int
      index of the atom 1
    a2: int
      index of the atom 2
    """
    bond: RDKitBond = self.datapoint.GetBondBetweenAtoms(a1 - 1, a2 - 1)
    if bond is None:
      return 0

    # get bond features
    f_bond: np.ndarray = np.asarray(bond_features(bond), dtype=float)

    self._extend_concat_feature(a1, f_bond)
    self._extend_concat_feature(a2, f_bond)
    return 1

  def _update_secondary_mappings(self, a1, a2):
    """
    Method to update `self.atom_to_incoming_bonds`, `self.bond_to_ini_atom` and `self.b2revb`
    with respect to the bond between atoms `a1` and `a2`.

    Parameters
    ----------
    a1: int
      index of the atom 1
    a2: int
      index of the atom 2
    """
    b1: int = self.num_bonds + 1  # bond index
    b2: int = self.num_bonds + 2  # reverse bond index

    self.atom_to_incoming_bonds[a2].append(b1)  # b1 = a1 --> 'a2'
    self.atom_to_incoming_bonds[a1].append(b2)  # b2 = a2 --> 'a1'

    self.bond_to_ini_atom.append(a1)  # b1 starts at a1
    self.bond_to_ini_atom.append(a2)  # b2 starts at a2 (remember, b2 =  b1+1)

    self.b2revb.append(b2)  # reverse bond of b1 is b2
    self.b2revb.append(b1)  # reverse bond of b2 is b1

  def _modify_based_on_max_bonds(self):
    """
    Method to make number of incoming bonds equal to maximum number of bonds.
    This is done by appending zeros to fill remaining space at each atom indices.
    """
    max_num_bonds: int = max(
        1,
        max(
            len(incoming_bonds)
            for incoming_bonds in self.atom_to_incoming_bonds))
    self.atom_to_incoming_bonds = [
        self.atom_to_incoming_bonds[a] + [0] *
        (max_num_bonds - len(self.atom_to_incoming_bonds[a]))
        for a in range(self.num_atoms + 1)
    ]

  def _replace_rev_bonds(self):
    """
    Method to replace the reverse bond indices with zeros.
    """
    for count, i in enumerate(self.b2revb):
      self.mapping[count][np.where(self.mapping[count] == i)] = 0


def generate_global_features(mol: RDKitMol,
                             features_generators: List[str]) -> np.ndarray:
  """
  Helper function for generating global features for a RDKit mol based on the given list of feature generators to be used.

  Parameters
  ----------
  mol: RDKitMol
    RDKit molecule to be featurized
  features_generators: List[str]
    List of names of the feature generators to be used featurization

  Returns
  -------
  global_features_array: np.ndarray
    Array of global features

  Examples
  --------
  >>> from rdkit import Chem
  >>> import deepchem as dc
  >>> mol = Chem.MolFromSmiles('C')
  >>> features_generators = ['morgan']
  >>> global_features = dc.feat.molecule_featurizers.dmpnn_featurizer.generate_global_features(mol, features_generators)
  >>> type(global_features)
  <class 'numpy.ndarray'>
  >>> len(global_features)
  2048
  >>> nonzero_features_indices = global_features.nonzero()[0]
  >>> nonzero_features_indices
  array([1264])
  >>> global_features[nonzero_features_indices[0]]
  1.0
  """
  global_features: List[np.ndarray] = []
  available_generators = GraphConvConstants.FEATURE_GENERATORS

  for generator in features_generators:
    if generator in available_generators:
      global_featurizer = available_generators[generator]
      if mol.GetNumHeavyAtoms() > 0:
        global_features.extend(global_featurizer.featurize(mol)[0])
      # for H2
      elif mol.GetNumHeavyAtoms() == 0:
        # not all features are equally long, so used methane as dummy molecule to determine length
        global_features.extend(
            np.zeros(
                len(global_featurizer.featurize(Chem.MolFromSmiles('C'))[0])))
    else:
      logger.warning(f"{generator} generator is not available in DMPNN")

  global_features_array: np.ndarray = np.asarray(global_features)

  # Fix nans in features
  replace_token = 0
  global_features_array = np.where(np.isnan(global_features_array),
                                   replace_token, global_features_array)

  return global_features_array


class DMPNNFeaturizer(MolecularFeaturizer):
  """
  This class is a featurizer for Directed Message Passing Neural Network (D-MPNN) implementation

  The default node(atom) and edge(bond) representations are based on
  `Analyzing Learned Molecular Representations for Property Prediction paper <https://arxiv.org/pdf/1904.01561.pdf>`_.

  The default node representation are constructed by concatenating the following values,
  and the feature length is 133.

  - Atomic num: A one-hot vector of this atom, in a range of first 100 atoms.
  - Degree: A one-hot vector of the degree (0-5) of this atom.
  - Formal charge: Integer electronic charge, -1, -2, 1, 2, 0.
  - Chirality: A one-hot vector of the chirality tag (0-3) of this atom.
  - Number of Hydrogens: A one-hot vector of the number of hydrogens (0-4) that this atom connected.
  - Hybridization: A one-hot vector of "SP", "SP2", "SP3", "SP3D", "SP3D2".
  - Aromatic: A one-hot vector of whether the atom belongs to an aromatic ring.
  - Mass: Atomic mass * 0.01

  The default edge representation are constructed by concatenating the following values,
  and the feature length is 14.

  - Bond type: A one-hot vector of the bond type, "single", "double", "triple", or "aromatic".
  - Same ring: A one-hot vector of whether the atoms in the pair are in the same ring.
  - Conjugated: A one-hot vector of whether this bond is conjugated or not.
  - Stereo: A one-hot vector of the stereo configuration (0-5) of a bond.

  If you want to know more details about features, please check the paper [1]_ and
  utilities in deepchem.utils.molecule_feature_utils.py.

  Examples
  --------
  >>> smiles = ["C1=CC=CN=C1", "C1CCC1"]
  >>> featurizer = DMPNNFeaturizer()
  >>> out = featurizer.featurize(smiles)
  >>> type(out[0])
  <class 'deepchem.feat.graph_data.GraphData'>
  >>> out[0].num_nodes
  6
  >>> out[0].num_node_features
  133
  >>> out[0].node_features.shape
  (6, 133)
  >>> out[0].num_edge_features
  14
  >>> out[0].num_edges
  12
  >>> out[0].edge_features.shape
  (12, 14)

  References
  ----------
  .. [1] Kearnes, Steven, et al. "Molecular graph convolutions: moving beyond fingerprints."
     Journal of computer-aided molecular design 30.8 (2016):595-608.

  Note
  ----
  This class requires RDKit to be installed.
  """

  def __init__(self,
               features_generators: Optional[List[str]] = None,
               is_adding_hs: bool = False,
               use_original_atom_ranks: bool = False):
    """
    Parameters
    ----------
    features_generator: List[str], default None
      List of global feature generators to be used.
    is_adding_hs: bool, default False
      Whether to add Hs or not.
    use_original_atom_ranks: bool, default False
      Whether to use original atom mapping or canonical atom mapping
    """
    self.features_generators = features_generators
    self.is_adding_hs = is_adding_hs
    super().__init__(use_original_atom_ranks)

  def _construct_bond_index(self, datapoint: RDKitMol) -> np.ndarray:
    """
    Construct edge (bond) index

    Parameters
    ----------
    datapoint: RDKitMol
      RDKit mol object.

    Returns
    -------
    edge_index: np.ndarray
      Edge (Bond) index
    """
    src: List[int] = []
    dest: List[int] = []
    for bond in datapoint.GetBonds():
      # add edge list considering a directed graph
      start, end = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
      src += [start, end]
      dest += [end, start]
    return np.asarray([src, dest], dtype=int)

  def _get_bond_features(self, datapoint: RDKitMol) -> np.ndarray:
    """
    Construct bond(edge) features for the given datapoint

    For each bond index, 2 bond feature arrays are added to the main features array,
    for the current bond and its reverse bond respectively.

    Note: This method of generating bond features ensures that the shape of the bond features array
          is always equal to (number of bonds, number of bond features), even if the number of bonds
          is equal to 0.

    Parameters
    ----------
    datapoint: RDKitMol
      RDKit mol object.

    Returns
    -------
    f_bonds: np.ndarray
      Bond features array
    """
    bonds: Chem.rdchem._ROBondSeq = datapoint.GetBonds()

    bond_fdim: int = GraphConvConstants.BOND_FDIM
    number_of_bonds: int = len(
        bonds) * 2  # Note the value is doubled to account for reverse bonds
    f_bonds: np.ndarray = np.empty((number_of_bonds, bond_fdim))

    for index in range(0, number_of_bonds, 2):
      bond_id: int = index // 2
      bond_feature: np.ndarray = np.asarray(bond_features(bonds[bond_id]),
                                            dtype=float)
      f_bonds[index] = bond_feature  # bond
      f_bonds[index + 1] = bond_feature  # reverse bond
    return f_bonds

  def _featurize(self, datapoint: RDKitMol, **kwargs) -> GraphData:
    """
    Calculate molecule graph features from RDKit mol object.

    Parameters
    ----------
    datapoint: RDKitMol
      RDKit mol object.

    Returns
    -------
    graph: GraphData
      A molecule graph object with features:
      - node_features: Node feature matrix with shape [num_nodes, num_node_features]
      - edge_index: Graph connectivity in COO format with shape [2, num_edges]
      - edge_features: Edge feature matrix with shape [num_edges, num_edge_features]
      - global_features: Array of global molecular features
    """
    if isinstance(datapoint, Chem.rdchem.Mol):
      if self.is_adding_hs:
        datapoint = Chem.AddHs(datapoint)
    else:
      raise ValueError(
          "Feature field should contain smiles for DMPNN featurizer!")

    # get atom features
    f_atoms: np.ndarray = np.asarray(
        [atom_features(atom) for atom in datapoint.GetAtoms()], dtype=float)

    # get edge(bond) features
    f_bonds: np.ndarray = self._get_bond_features(datapoint)

    # get edge index
    edge_index: np.ndarray = self._construct_bond_index(datapoint)

    # get global features
    global_features: np.ndarray = np.empty(0)
    if self.features_generators is not None:
      global_features = generate_global_features(datapoint,
                                                 self.features_generators)

    return GraphData(node_features=f_atoms,
                     edge_index=edge_index,
                     edge_features=f_bonds,
                     global_features=global_features)



"""
Data Structures used to represented molecules for convolutions.
"""
# flake8: noqa

import csv
import random
import numpy as np


def cumulative_sum_minus_last(l, offset=0):
  """Returns cumulative sums for set of counts, removing last entry.

  Returns the cumulative sums for a set of counts with the first returned value
  starting at 0. I.e [3,2,4] -> [0, 3, 5]. Note last sum element 9 is missing.
  Useful for reindexing

  Parameters
  ----------
  l: list
    List of integers. Typically small counts.
  """
  return np.delete(np.insert(np.cumsum(l, dtype=np.int32), 0, 0), -1) + offset


def cumulative_sum(l, offset=0):
  """Returns cumulative sums for set of counts.

  Returns the cumulative sums for a set of counts with the first returned value
  starting at 0. I.e [3,2,4] -> [0, 3, 5, 9]. Keeps final sum for searching.
  Useful for reindexing.

  Parameters
  ----------
  l: list
    List of integers. Typically small counts.
  """
  return np.insert(np.cumsum(l), 0, 0) + offset


class ConvMol(object):
  """Holds information about a molecules.

  Resorts order of atoms internally to be in order of increasing degree. Note
  that only heavy atoms (hydrogens excluded) are considered here.
  """

  def __init__(self, atom_features, adj_list, max_deg=10, min_deg=0):
    """
    Parameters
    ----------
    atom_features: np.ndarray
      Has shape (n_atoms, n_feat)
    adj_list: list
      List of length n_atoms, with neighor indices of each atom.
    max_deg: int, optional
      Maximum degree of any atom.
    min_deg: int, optional
      Minimum degree of any atom.
    """

    self.atom_features = atom_features
    self.n_atoms, self.n_feat = atom_features.shape
    self.deg_list = np.array([len(nbrs) for nbrs in adj_list], dtype=np.int32)
    self.canon_adj_list = adj_list
    self.deg_adj_lists = []
    self.deg_slice = []
    self.max_deg = max_deg
    self.min_deg = min_deg

    self.membership = self.get_num_atoms() * [0]

    self._deg_sort()

    # Get the degree id list (which corrects for min_deg)
    self.deg_id_list = np.array(self.deg_list) - min_deg

    # Get the size of each degree block
    deg_size = [
        self.get_num_atoms_with_deg(deg)
        for deg in range(self.min_deg, self.max_deg + 1)
    ]

    self.degree_list = []
    for i, deg in enumerate(range(self.min_deg, self.max_deg + 1)):
      self.degree_list.extend([deg] * deg_size[i])

    # Get the the start indices for items in each block
    self.deg_start = cumulative_sum(deg_size)

    # Get the node indices when they are reset when the degree changes
    deg_block_indices = [
        i - self.deg_start[self.deg_list[i]] for i in range(self.n_atoms)
    ]

    # Convert to numpy array
    self.deg_block_indices = np.array(deg_block_indices, dtype=np.int32)

  def get_atoms_with_deg(self, deg):
    """Retrieves atom_features with the specific degree"""
    start_ind = self.deg_slice[deg - self.min_deg, 0]
    size = self.deg_slice[deg - self.min_deg, 1]
    return self.atom_features[start_ind:(start_ind + size), :]

  def get_num_atoms_with_deg(self, deg):
    """Returns the number of atoms with the given degree"""
    return self.deg_slice[deg - self.min_deg, 1]

  def get_num_atoms(self):
    return self.n_atoms

  def _deg_sort(self):
    """Sorts atoms by degree and reorders internal data structures.

    Sort the order of the atom_features by degree, maintaining original order
    whenever two atom_features have the same degree.
    """
    old_ind = range(self.get_num_atoms())
    deg_list = self.deg_list
    new_ind = list(np.lexsort((old_ind, deg_list)))

    num_atoms = self.get_num_atoms()

    # Reorder old atom_features
    self.atom_features = self.atom_features[new_ind, :]

    # Reorder old deg lists
    self.deg_list = [self.deg_list[i] for i in new_ind]

    # Sort membership
    self.membership = [self.membership[i] for i in new_ind]

    # Create old to new dictionary. not exactly intuitive
    old_to_new = dict(zip(new_ind, old_ind))

    # Reorder adjacency lists
    self.canon_adj_list = [self.canon_adj_list[i] for i in new_ind]
    self.canon_adj_list = [[old_to_new[k]
                            for k in self.canon_adj_list[i]]
                           for i in range(len(new_ind))]

    # Get numpy version of degree list for indexing
    deg_array = np.array(self.deg_list)

    # Initialize adj_lists, which supports min_deg = 1 only
    self.deg_adj_lists = (self.max_deg + 1 - self.min_deg) * [0]

    # Parse as deg separated
    for deg in range(self.min_deg, self.max_deg + 1):
      # Get indices corresponding to the current degree
      rng = np.array(range(num_atoms))
      indices = rng[deg_array == deg]

      # Extract and save adjacency list for the current degree
      to_cat = [self.canon_adj_list[i] for i in indices]
      if len(to_cat) > 0:
        adj_list = np.vstack([self.canon_adj_list[i] for i in indices])
        self.deg_adj_lists[deg - self.min_deg] = adj_list.astype(np.int32)

      else:
        self.deg_adj_lists[deg - self.min_deg] = np.zeros(
            [0, deg], dtype=np.int32)

    # Construct the slice information
    deg_slice = np.zeros([self.max_deg + 1 - self.min_deg, 2], dtype=np.int32)

    for deg in range(self.min_deg, self.max_deg + 1):
      if deg == 0:
        deg_size = np.sum(deg_array == deg)
      else:
        deg_size = self.deg_adj_lists[deg - self.min_deg].shape[0]

      deg_slice[deg - self.min_deg, 1] = deg_size
      # Get the cumulative indices after the first index
      if deg > self.min_deg:
        deg_slice[deg - self.min_deg, 0] = (
            deg_slice[deg - self.min_deg - 1, 0] +
            deg_slice[deg - self.min_deg - 1, 1])

    # Set indices with zero sized slices to zero to avoid indexing errors
    deg_slice[:, 0] *= (deg_slice[:, 1] != 0)
    self.deg_slice = deg_slice

  def get_atom_features(self):
    """Returns canonicalized version of atom features.

    Features are sorted by atom degree, with original order maintained when
    degrees are same.
    """
    return self.atom_features

  def get_adjacency_list(self):
    """Returns a canonicalized adjacency list.

    Canonicalized means that the atoms are re-ordered by degree.

    Returns
    -------
    list
      Canonicalized form of adjacency list.
    """
    return self.canon_adj_list

  def get_deg_adjacency_lists(self):
    """Returns adjacency lists grouped by atom degree.

    Returns
    -------
    list
      Has length (max_deg+1-min_deg). The element at position deg is
      itself a list of the neighbor-lists for atoms with degree deg.
    """
    return self.deg_adj_lists

  def get_deg_slice(self):
    """Returns degree-slice tensor.

    The deg_slice tensor allows indexing into a flattened version of the
    molecule's atoms. Assume atoms are sorted in order of degree. Then
    deg_slice[deg][0] is the starting position for atoms of degree deg in
    flattened list, and deg_slice[deg][1] is the number of atoms with degree deg.

    Note deg_slice has shape (max_deg+1-min_deg, 2).

    Returns
    -------
    deg_slice: np.ndarray
      Shape (max_deg+1-min_deg, 2)
    """
    return self.deg_slice

  # TODO(rbharath): Can this be removed?
  @staticmethod
  def get_null_mol(n_feat, max_deg=10, min_deg=0):
    """Constructs a null molecules

    Get one molecule with one atom of each degree, with all the atoms
    connected to themselves, and containing n_feat features.

    Parameters
    ----------
    n_feat : int
        number of features for the nodes in the null molecule
    """
    # Use random insted of zeros to prevent weird issues with summing to zero
    atom_features = np.random.uniform(0, 1, [max_deg + 1 - min_deg, n_feat])
    canon_adj_list = [
        deg * [deg - min_deg] for deg in range(min_deg, max_deg + 1)
    ]

    return ConvMol(atom_features, canon_adj_list)

  @staticmethod
  def agglomerate_mols(mols, max_deg=10, min_deg=0):
    """Concatenates list of ConvMol's into one mol object that can be used to feed
    into tensorflow placeholders. The indexing of the molecules are preseved during the
    combination, but the indexing of the atoms are greatly changed.

    Parameters
    ----
    mols: list
      ConvMol objects to be combined into one molecule."""

    num_mols = len(mols)

    # Combine the features, then sort them by (atom_degree, mol_index)
    atoms_by_deg = np.concatenate([x.atom_features for x in mols])
    degree_vector = np.concatenate([x.degree_list for x in mols], axis=0)
    # Mergesort is a "stable" sort, so the array maintains it's secondary sort of mol_index
    order = degree_vector.argsort(kind='mergesort')
    ordered = np.empty(order.shape, np.int32)
    ordered[order] = np.arange(order.shape[0], dtype=np.int32)
    all_atoms = atoms_by_deg[order]

    # Create a map from the original atom indices within each molecule to the
    # indices in the combined object.
    mol_atom_map = []
    index_start = 0
    for mol in mols:
      mol_atom_map.append(
          ordered[index_start:index_start + mol.get_num_atoms()])
      index_start += mol.get_num_atoms()

    # Sort all atoms by degree.
    # Get the size of each atom list separated by molecule id, then by degree
    mol_deg_sz = np.zeros([max_deg - min_deg + 1, num_mols], dtype=np.int32)
    for i, mol in enumerate(mols):
      mol_deg_sz[:, i] += mol.deg_slice[:, 1]

    # Get the final size of each degree block
    deg_sizes = np.sum(mol_deg_sz, axis=1)

    # Get the index at which each degree starts, not resetting after each degree
    # And not stopping at any specific molecule

    deg_start = cumulative_sum_minus_last(deg_sizes)

    # Get the tensorflow object required for slicing (deg x 2) matrix, with the
    # first column telling the start indices of each degree block and the
    # second colum telling the size of each degree block
    deg_slice = np.array(list(zip(deg_start, deg_sizes)))

    # Determine the membership (atom i belongs to molecule membership[i])
    membership = np.empty(all_atoms.shape[0], np.int32)
    for i in range(num_mols):
      membership[mol_atom_map[i]] = i

    # Initialize the new degree separated adjacency lists
    deg_adj_lists = [
        np.empty([deg_sizes[deg], deg], dtype=np.int32)
        for deg in range(min_deg, max_deg + 1)
    ]

    # Update the old adjacency lists with the new atom indices and then combine
    # all together
    for deg in range(min_deg, max_deg + 1):
      row = 0  # Initialize counter
      deg_id = deg - min_deg  # Get corresponding degree id

      # Iterate through all the molecules
      for mol_id in range(num_mols):
        # Get the adjacency lists for this molecule and current degree id
        nbr_list = mols[mol_id].deg_adj_lists[deg_id]

        # Correct all atom indices to the final indices, and then save the
        # results into the new adjacency lists
        if nbr_list.shape[0] > 0:
          if nbr_list.dtype == np.int32:
            final_id = mol_atom_map[mol_id][nbr_list]
            deg_adj_lists[deg_id][row:(row + nbr_list.shape[0])] = final_id
            row += nbr_list.shape[0]
          else:
            for i in range(nbr_list.shape[0]):
              for j in range(nbr_list.shape[1]):
                deg_adj_lists[deg_id][row, j] = mol_atom_map[mol_id][nbr_list[
                    i, j]]
              # Increment once row is done
              row += 1

    # Get the final aggregated molecule
    concat_mol = MultiConvMol(all_atoms, deg_adj_lists, deg_slice, membership,
                              num_mols)
    return concat_mol


class MultiConvMol(object):
  """Holds information about multiple molecules, for use in feeding information
     into tensorflow. Generated using the agglomerate_mols function
  """

  def __init__(self, nodes, deg_adj_lists, deg_slice, membership, num_mols):
    self.nodes = nodes
    self.deg_adj_lists = deg_adj_lists
    self.deg_slice = deg_slice
    self.membership = membership
    self.num_mols = num_mols
    self.num_atoms = nodes.shape[0]

  def get_deg_adjacency_lists(self):
    return self.deg_adj_lists

  def get_atom_features(self):
    return self.nodes

  def get_num_atoms(self):
    return self.num_atoms

  def get_num_molecules(self):
    return self.num_mols


class WeaveMol(object):
  """Molecular featurization object for weave convolutions.

  These objects are produced by WeaveFeaturizer, and feed into
  WeaveModel. The underlying implementation is inspired by [1]_.


  References
  ----------
  .. [1] Kearnes, Steven, et al. "Molecular graph convolutions: moving beyond fingerprints." Journal of computer-aided molecular design 30.8 (2016): 595-608.
  """

  def __init__(self, nodes, pairs, pair_edges):
    self.nodes = nodes
    self.pairs = pairs
    self.num_atoms = self.nodes.shape[0]
    self.n_features = self.nodes.shape[1]
    self.pair_edges = pair_edges

  def get_pair_edges(self):
    return self.pair_edges

  def get_pair_features(self):
    return self.pairs

  def get_atom_features(self):
    return self.nodes

  def get_num_atoms(self):
    return self.num_atoms

  def get_num_features(self):
    return self.n_features
